import ipydatawidgets
import ipywidgets
from ipywidgets import Layout, interact, interactive, fixed, interact_manual
import nbvv.img_prep
import numpy
import traitlets


@ipywidgets.register
class VolumeWidget(ipywidgets.DOMWidget):
    _view_name = traitlets.Unicode("VolumeWidgetView").tag(sync=True)

    _view_module = traitlets.Unicode("nbvv").tag(sync=True)

    _view_module_version = traitlets.Unicode("1.0.0").tag(sync=True)

    image = ipydatawidgets.NDArray(numpy.zeros(0, dtype=numpy.uint8)).tag(
        sync=True, **ipydatawidgets.array_serialization
    )

    size = traitlets.Tuple().tag(sync=True)

    dimensions = traitlets.Dict().tag(sync=True)

    metadata = traitlets.Dict({"foo": "bar"}).tag(sync=True)

    density = traitlets.Float(0.1).tag(sync=True)
    brightness = traitlets.Float(0.1).tag(sync=True)


# expect CZYX
def volshow(
    image, spacing=(1.0, 1.0, 1.0), density=0.1, brightness=1.0, channel_names=None
):
    # assume CZYX if 4d and ZYX if 3d.
    if len(image.shape) > 4:
        return "Image must be 3 or 4 dimensional"
    if len(image.shape) < 3:
        return "Image must be 3 or 4 dimensional"
    # add a channel dimension if needed
    if len(image.shape) == 3:
        image = numpy.expand_dims(image, axis=0)

    volume_widget = VolumeWidget()

    dims_object = nbvv.img_prep.atlas_dimensions(
        image, physical_pixel_size=spacing, channel_names=channel_names
    )
    # image MUST have a name
    dims_object["name"] = "Image0"
    volume_widget.dimensions = dims_object

    # downsample and normalize for browser rendering
    image = nbvv.img_prep.img_prep(
        image,  # CZYX
        shape=(
            volume_widget.dimensions["tile_width"],
            volume_widget.dimensions["tile_height"],
        ),
    )

    # pass to javascript as array of ZYX volumes, one per channel
    volume_widget.image = [image[index] for index in range(image.shape[0])]
    return volume_widget



from ._version import version_info, __version__

from .volume import *


def _prefix():
    import sys
    from pathlib import Path

    prefix = sys.prefix
    here = Path(__file__).parent
    # for when in dev mode
    if (here.parent / "share/jupyter/nbextensions/nbvv").exists():
        prefix = here.parent
    return prefix


def _jupyter_labextension_paths():
    return [{"src": f"{_prefix()}/share/jupyter/labextensions/nbvv/", "dest": "nbvv",}]


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": f"{_prefix()}/share/jupyter/nbextensions/nbvv/",
            "dest": "nbvv",
            "require": "nbvv/extension",
        }
    ]




import numpy

import nbvv.volume


class TestVolume:
    def test_image(self):
        widget = nbvv.volume.VolumeWidget()

        assert widget.image.size == 0

    def test_metadata(self):
        widget = nbvv.volume.VolumeWidget()

        assert widget.metadata == {"foo": "bar"}


def test_volshow():
    image = numpy.random.random((10, 256, 256, 3))

    widget = nbvv.volume.volshow(image)

    numpy.testing.assert_array_equal(widget.image, image)



import math
import numpy
import scipy.ndimage
import skimage.transform
import scipy.misc

__all__ = ["img_prep"]


def resize(image, output_shape):
    # YX
    scale = (
        float(output_shape[1]) / float(image.shape[1]),
        float(output_shape[0]) / float(image.shape[2]),
    )

    # ZYX
    response = numpy.zeros((image.shape[0], output_shape[1], output_shape[0]))

    for index, image_plane in enumerate(image):
        # order=0 nearest neighbor for labeled/segmentation images,
        # order>0 for raw intensity data? 1 is default
        response[index] = skimage.transform.rescale(
            image_plane, scale, preserve_range=True, order=0, anti_aliasing=False
        )
        # response[index] = skimage.transform.resize(
        #     image_plane, output_shape, mode="reflect", anti_aliasing=downsampling
        # )

    return response


def atlas_dimensions(
    aics_image, max_edge=2048, channel_names=None, physical_pixel_size=(1.0, 1.0, 1.0)
):
    # incoming is CZYX
    tile_width, tile_height, stack_height = (
        aics_image.shape[3],
        aics_image.shape[2],
        aics_image.shape[1],
    )
    # maintain aspect ratio of images
    # initialize atlas with one row of all slices
    atlas_width = tile_width * stack_height
    atlas_height = tile_height
    ratio = float(atlas_width) / float(atlas_height)

    # these next steps attempt to optimize the atlas into a square shape
    # TODO - there must be a way to do this with a single calculation
    for r in range(2, stack_height):
        new_rows = math.ceil(float(stack_height) / r)
        adjusted_width = int(tile_width * new_rows)
        adjusted_height = int(tile_height * r)
        new_ratio = float(max(adjusted_width, adjusted_height)) / float(
            min(adjusted_width, adjusted_height)
        )
        if new_ratio < ratio:
            ratio = new_ratio
            atlas_width = adjusted_width
            atlas_height = adjusted_height
        else:
            # we've found the rows and columns that make this the most square image
            break

    cols = int(atlas_width // tile_width)
    rows = int(atlas_height // tile_height)

    if max_edge < atlas_width or max_edge < atlas_height:
        tile_width = math.floor(max_edge / cols)
        tile_height = math.floor(max_edge / rows)
        atlas_width = tile_width * cols
        atlas_height = tile_height * rows

    dims = {
        "tile_width": int(tile_width),
        "tile_height": int(tile_height),
        "rows": int(rows),
        "cols": int(cols),
        "atlas_width": int(atlas_width),
        "atlas_height": int(atlas_height),
        "width": aics_image.shape[3],
        "height": aics_image.shape[2],
        "channels": aics_image.shape[0],
        "tiles": aics_image.shape[1],
    }

    if channel_names is not None:
        dims["channel_names"] = channel_names
    else:
        dims["channel_names"] = ["CH_" + str(i) for i in range(aics_image.shape[0])]

    if physical_pixel_size is not None:
        dims["pixel_size_x"] = physical_pixel_size[0]
        dims["pixel_size_y"] = physical_pixel_size[1]
        dims["pixel_size_z"] = physical_pixel_size[2]
    else:
        dims["pixel_size_x"] = 1
        dims["pixel_size_y"] = 1
        dims["pixel_size_z"] = 1

    return dims


def img_prep(img, shape=(128, 128)):
    """Given an input 4D CZYX image, prep for display

    Parameters
    ----------
    img : array
        Numpy array in CZYX (channel, plane, row, column) format
    shape : tuple(int)
        This is the volume's preferred xy size for texture atlassing for realtime display
        as computed by atlas_dimemsions

    Returns
    -------
    img : list of arrays
        one channel to each entry in the list
    """

    # C Z Y X
    channels = img.shape[0]

    # C Z Y X
    response = numpy.zeros(
        (channels, img.shape[1], shape[1], shape[0]), dtype=numpy.uint8
    )

    for channel_index in range(channels):
        channel_data = img[channel_index]
        channel_data = channel_data.astype(numpy.float32)

        resized_channel = resize(channel_data, (shape[0], shape[1]))

        mn = min(0, resized_channel.min())
        mx = resized_channel.max()
        resized_channel = 255.0 * (resized_channel - mn) / (mx - mn)
        # atlas = np.interp(atlas, (min(0, atlas.min()), atlas.max()), (0.0, 255.0))
        resized_channel = resized_channel.astype(numpy.uint8)

        response[channel_index] = resized_channel

    return response



from __future__ import print_function
from setuptools import setup, find_packages, Command
from setuptools.command.sdist import sdist
from setuptools.command.build_py import build_py
from setuptools.command.egg_info import egg_info
from subprocess import check_call
import os
from os.path import join as pjoin
from pathlib import Path
import sys
import platform
from distutils import log
from jupyter_packaging import (
    create_cmdclass,
    install_npm,
    ensure_targets,
    combine_commands,
    get_version,
)

here = os.path.dirname(os.path.abspath(__file__))

log.set_verbosity(log.DEBUG)


def skip_if_exists(paths, CommandClass):
    """Skip a command if list of paths exists."""

    def should_skip():
        return any(not Path(path).exist() for path in paths)

    class SkipIfExistCommand(Command):
        def initialize_options(self):
            if not should_skip:
                self.command = CommandClass(self.distribution)
                self.command.initialize_options()
            else:
                self.command = None

        def finalize_options(self):
            if self.command is not None:
                self.command.finalize_options()

        def run(self):
            if self.command is not None:
                self.command.run()

    return SkipIfExistCommand


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


name = "nbvv"
LONG_DESCRIPTION = read("README.md")
version = get_version(pjoin(name, "_version.py"))
js_dir = pjoin(here, "js")
jstargets = [
    pjoin(js_dir, "lib", "index.js"),
    pjoin("share", "jupyter", "nbextensions", "nbvv", "index.js"),
]
data_files_spec = [
    ("share/jupyter/nbextensions/nbvv", "share/jupyter/nbextensions/nbvv", "*.js"),
    ("share/jupyter/labextensions/nbvv", "share/jupyter/labextensions/nbvv", "*"),
    (
        "share/jupyter/labextensions/nbvv/static",
        "share/jupyter/labextensions/nbvv/static",
        "*",
    ),
    ("etc/jupyter/nbconfig/notebook.d", "etc/jupyter/nbconfig/notebook.d", "nbvv.json"),
]

js_command = combine_commands(
    install_npm(js_dir, build_cmd="build"), ensure_targets(jstargets),
)

cmdclass = create_cmdclass("jsdeps", data_files_spec=data_files_spec)
is_repo = os.path.exists(os.path.join(here, ".git"))
if is_repo:
    cmdclass["jsdeps"] = js_command
else:
    cmdclass["jsdeps"] = skip_if_exists(jstargets, js_command)


setup(
    name=name,
    version=version,
    description="Interactive volumetric voxel viewing",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    include_package_data=True,
    cmdclass=cmdclass,
    install_requires=[
        "ipydatawidgets>=4.3.1",
        "ipywidgets>=7.7.1",
        "jupyter>=1.0.0",
        "jupyterlab>=3.4.4",
        "jupyter-packaging>=0.12.2",
        "numpy>=1.14.3",
        "scikit-image>=0.13.0",
        "scipy>=1.0.0",
    ],
    license="MIT",
    packages=find_packages(),
    zip_safe=False,
    author="AICS",
    author_email="danielt@alleninstitute.org",
    url="https://github.com/allen-cell-animated/nbvv",
    keywords=["ipython", "jupyter", "widgets", "volume rendering"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: IPython",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Graphics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)
